"""
https://raw.githubusercontent.com/EMNLP2019LSAN/LSAN/master/attention/model.py
"""
import torch
import torch.nn.functional as F
from .abstracts import TextClassificationAbstract
from ..representation import get

class LSANOriginal(TextClassificationAbstract):
    """
    https://raw.githubusercontent.com/EMNLP2019LSAN/LSAN/master/attention/model.py
    """
    def __init__(self, classes, representation, label_embed=None, label_freeze=True, lstm_hid_dim=300, d_a=200 ,max_len=500,**kwargs):
        super(LSANOriginal, self).__init__(**kwargs)
        #My Stuff
        self.max_len = max_len


        # Original
        self.n_classes = len(classes)
        self.embedding, self.tokenizer = get(representation, freeze=True)
        self.embedding_dim = self.embedding(torch.LongTensor([[0]])).shape[-1]

        if label_embed is not None:
            self.label_embed = torch.nn.Embedding(label_embed.shape[0], label_embed.shape[1])
            self.label_embed.from_pretrained(torch.FloatTensor(label_embed), freeze=label_freeze)
        else:
            self.label_embed = torch.nn.Embedding(self.n_classes, self.embedding_dim)

        self.lstm = torch.nn.LSTM(self.embedding_dim, hidden_size=lstm_hid_dim, num_layers=1,
                                  batch_first=True, bidirectional=True)

        self.linear_first = torch.nn.Linear(lstm_hid_dim * 2, d_a)
        self.linear_second = torch.nn.Linear(d_a, self.n_classes)

        self.weight1 = torch.nn.Linear(lstm_hid_dim * 2, 1)
        self.weight2 = torch.nn.Linear(lstm_hid_dim * 2, 1)

        self.output_layer = torch.nn.Linear(lstm_hid_dim * 2, self.n_classes)
        self.embedding_dropout = torch.nn.Dropout(p=0.5)
        self.lstm_hid_dim = lstm_hid_dim
        self.build()

    def init_hidden(self, size):
        return (torch.randn(2, size, self.lstm_hid_dim).to(self.device),
                torch.randn(2, size, self.lstm_hid_dim).to(self.device))

    def forward(self, x):
        embeddings = self.embedding(x)
        embeddings = self.embedding_dropout(embeddings)
        # step1 get LSTM outputs
        # hidden_state = self.init_hidden(x.shape[0])
        outputs, hidden_state = self.lstm(embeddings)#, hidden_state)
        # step2 get self-attention
        selfatt = torch.tanh(self.linear_first(outputs))
        selfatt = self.linear_second(selfatt)
        selfatt = F.softmax(selfatt, dim=1)
        selfatt = selfatt.transpose(1, 2)
        self_att = torch.bmm(selfatt, outputs)
        # step3 get label-attention
        h1 = outputs[:, :, :self.lstm_hid_dim]
        h2 = outputs[:, :, self.lstm_hid_dim:]

        label = self.embedding_dropout(self.label_embed.weight.data)
        m1 = torch.bmm(label.expand(x.shape[0], self.n_classes, self.lstm_hid_dim), h1.transpose(1, 2))
        m2 = torch.bmm(label.expand(x.shape[0], self.n_classes, self.lstm_hid_dim), h2.transpose(1, 2))
        label_att = torch.cat((torch.bmm(m1, h1), torch.bmm(m2, h2)), 2)
        # label_att = F.normalize(label_att, p=2, dim=-1)
        # self_att = F.normalize(self_att, p=2, dim=-1) #all can
        weight1 = torch.sigmoid(self.weight1(label_att))
        weight2 = torch.sigmoid(self.weight2(self_att))
        weight1 = weight1 / (weight1 + weight2)
        weight2 = 1 - weight1

        doc = weight1 * label_att + weight2 * self_att
        # there two method, for simple, just add
        # also can use linear to do it
        doc = self.embedding_dropout(doc)
        avg_sentence_embeddings = torch.sum(doc, 1) / self.n_classes

        pred = self.output_layer(avg_sentence_embeddings)
        return pred



"""
Submodule containing implementations of neural network models for Multilabel Architectures.
"""

# Well Tested
from .XMLCNN import XMLCNN
from .KimCNN import KimCNN

#LSAN Variants
from .LSAN import LSANOriginal
from .LSANOriginalTransformer import LSANOriginalTransformer


# Geometric Models
try:
    from .ZAGCNN import ZAGCNN
    from .ZAGCNNLM import ZAGCNNLM
    from .SKG_ML import SKG
except:
    print("pytorch_geometric not installed.")
    pass

from .LSANOriginalTransformerNoClasses import LSANOriginalTransformerNoClasses
from .lm_mogrifier import MogrifierLMCharacter, MogrifierLMWord


"""
https://raw.githubusercontent.com/EMNLP2019LSAN/LSAN/master/attention/model.py
"""
import torch
import torch.nn.functional as F
from .abstracts import TextClassificationAbstract
from ..representation import get
import re


class LSANOriginalTransformerNoClasses(TextClassificationAbstract):
    """
    https://raw.githubusercontent.com/EMNLP2019LSAN/LSAN/master/attention/model.py
    """
    def __init__(self, classes, method="glove", scale="mean", norm=False, representation="roberta", use_lstm=True, d_a=200, max_len=400, n_layers=4, **kwargs):
        super(LSANOriginalTransformerNoClasses, self).__init__(**kwargs)
        #My Stuff
        self.classes = classes
        self.max_len = max_len
        self.use_lstm = use_lstm
        self.n_layers = 4
        self.representation=representation
        self.method = method
        self.scale = scale
        self.norm = norm
        self.n_layers=n_layers

        # Original
        self.n_classes = len(classes)
        self.representation = representation
        self._init_input_representations()

        self.create_labels(classes, method=self.method, scale=self.scale)

        if use_lstm:
            self.lstm = torch.nn.LSTM(self.embeddings_dim, hidden_size=self.label_embedding_dim, num_layers=1,
                                      batch_first=True, bidirectional=True)
        else:
            self.lstm = torch.nn.Linear(self.embeddings_dim, self.label_embedding_dim * 2)

        self.linear_first = torch.nn.Linear(self.label_embedding_dim * 2, d_a)
        self.linear_second = torch.nn.Linear(self.label_embedding_dim , d_a)

        self.weight1 = torch.nn.Linear(self.label_embedding_dim * 2, 1)
        self.weight2 = torch.nn.Linear(self.label_embedding_dim * 2, 1)

        self.output_layer = torch.nn.Linear(self.label_embedding_dim * 2, 1)
        self.embedding_dropout = torch.nn.Dropout(p=0.5)

        # self.connection = torch.nn.Linear(self.label_embedding_dim * 2, 1)

        self.build()

    def init_hidden(self, size):
        return (torch.randn(2, size, self.label_embedding_dim).to(self.device),
                torch.randn(2, size, self.label_embedding_dim).to(self.device))

    def forward(self, x):
        if self.n_layers == 1:
            with torch.no_grad():
                embeddings = self.embedding(x)[0]
        else:
            with torch.no_grad():
                embeddings = torch.cat(self.embedding(x)[2][self.n_layers:], -1)

        embeddings = self.embedding_dropout(embeddings)
        # step1 get LSTM outputs
        # hidden_state = self.init_hidden(x.shape[0])
        outputs = self.lstm(embeddings)#, hidden_state)
        if self.use_lstm:
            outputs = outputs[0]
        # step2 get self-attention
        selfatt = torch.tanh(self.linear_first(outputs))
        selfatt = torch.matmul(selfatt, self.linear_second(self.label_embedding).t())
        selfatt = F.softmax(selfatt, dim=1)
        selfatt = selfatt.transpose(1, 2)
        self_att = torch.bmm(selfatt, outputs)
        # step3 get label-attention

        h1 = outputs[:, :, :self.label_embedding_dim]
        h2 = outputs[:, :, self.label_embedding_dim:]

        label = self.label_embedding
        m1 = torch.bmm(label.expand(x.shape[0], *label.shape), h1.transpose(1, 2))
        m2 = torch.bmm(label.expand(x.shape[0], *label.shape), h2.transpose(1, 2))
        label_att = torch.relu(torch.cat((torch.bmm(m1, h1), torch.bmm(m2, h2)), 2))
        # label_att = self.connection(label_att.transpose(-1,-2)).transpose(-1,-2)


        # label_att = F.normalize(label_att, p=2, dim=-1)
        # self_att = F.normalize(self_att, p=2, dim=-1) #all can
        weight1 = torch.sigmoid(self.weight1(label_att))
        weight2 = torch.sigmoid(self.weight2(self_att))


        weight1 = weight1 / (weight1 + weight2)
        weight2 = 1 - weight1

        doc = weight1 * label_att + weight2 * self_att
        # there two method, for simple, just add
        # also can use linear to do it
        doc = self.embedding_dropout(doc)
         # = torch.sum(doc, -1)

        pred = self.output_layer(doc / self.label_embedding_dim).squeeze()
        return pred

    def create_labels(self, classes, method="repeat",scale="mean"):
        assert method in ("repeat","generate","embed", "glove"), 'method has to be one of ("repeat","generate","embed", "glove")'
        self.classes = classes
        if method=="repeat":
            from ..representation import get_lm_repeated
            l = get_lm_repeated(self.classes, self.representation)
        if method == "generate":
            from ..representation import get_lm_generated
            l = get_lm_generated(self.classes, self.representation)
        if method == "embed":
            l = self.embedding(self.tokenizer(self.classes.keys()).to(list(self.parameters())[0].device))[1]
        if method == "glove":
            from ..representation import get_word_embedding_mean
            l = get_word_embedding_mean(
                [" ".join(re.split("[/ _-]", x.lower())) for x in self.classes.keys()],
                "glove300")

        if scale=="mean":
            print("subtracting mean")
            l = l - l.mean(0,keepdim=True)
        if scale == "normalize":
            print("normalizing")
            l = l/l.norm(p=2,dim=-1,keepdim=True)

        self.label_embedding = torch.nn.Parameter(l.to(self.device))
        self.label_embedding.requires_grad = False
        self.label_embedding_dim = self.label_embedding.shape[-1]



from typing import List

from torch.utils.data import Dataset, DataLoader
from tqdm import tqdm

from ..representation import get, is_transformer
import torch

class _temporal_dataset(Dataset):
    def __init__(self, l):
        self.l = l
    def __len__(self):
        return len(self.l)
    def __getitem__(self, item):
        return self.l[item]

class Embedder:
    """
    A class for embedding lists of sentences.
    """
    def __init__(self, representation, method="first_token", device="cpu", return_device="cpu"):
        """
        Holding an embedding to embed lists of sentences with huggingface or glove.

        This class returns an embedding vector for every whitespace separated token in the data.
        (Tested only on all BERT-like models which use a prefixe for whitespaces.)
        For now only the first_token policy is supported.

        ToDo:
            - Support Pooling method
            - Averaging Method

        :param representation: Name of the representation (language model or glove vectors)
        :param device: The device to use for the computation of the embeddings
        :param return_device: The device to store the results on ( For large data it is adviced to leave this on cpu, for smaller amount of data that will be used for further computation
        it can be useful to leave it on the device for further processing
        """
        self.representation = representation
        self.emb, self.tok = get(self.representation)
        self.device = device
        self.emb = self.emb.to(self.device)
        self.return_device = return_device
        assert method in ("first_token",)
        self.method = method

    def embed(self, sentences: List, pad=None):
        """
        Embedding method for a list of sentences.
        :param sentences:  List of sentences
        :param pad: (default: None) If pad is set all sentences will be padded (or cut repectively) to the desired length.
        :return: if pad is None a list of embeddings (with varying lengths) is returned, is pad is set a tensor of (num_sentences, pad, embedding_size) will be returned.
        """

        if is_transformer(self.representation):
            t, ind = self.tok(sentences, return_start=True)
            with torch.no_grad():
                embeddings = self.emb(t.to(self.device))[0].to(self.return_device)
            embeddings = [e[i]for e, i in zip(embeddings, ind)]
            if pad is not None:
                r = torch.zeros((len(embeddings), pad, embeddings[0].shape[-1]))
                for i, e in enumerate(embeddings):
                    r[i, :min(e.shape[0] - 1, pad), :] = e[min(e.shape[0] - 1, pad)]
                embeddings = r
        else:
            s_length = [len(x.split()) for x in sentences]
            t = self.tok(sentences, pad if pad is not None else max(s_length))
            with torch.no_grad():
                embeddings = self.emb(t.to(self.device)).to(self.return_device)
            if pad is None:
                embeddings = [e[:i] for e,i in zip(embeddings,s_length)]

        return embeddings

    def embed_batch(self, sentences: List, batch_size=64, pad = None):
        """
        Embed large data batchwise.

        Same as embed method but computing the data batchwise to prevent OOM.

        :param sentences: List of sentences.
        :param batch_size: Size of the batches
        :param pad: (default: None) If pad is set all sentences will be padded (or cut repectively) to the desired length.
        :return: if pad is None a list of embeddings (with varying lengths) is returned, is pad is set a tensor of (num_sentences, pad, embedding_size) will be returned.

        """
        ds = _temporal_dataset(sentences)
        r = []
        for b in tqdm(DataLoader(ds, batch_size=batch_size)):
            r.append(self.embed(b, pad=pad))
        if pad is not None:
            from torch import cat
            r = cat(r,0)
        return r

    def embed_batch_iterator(self, sentences: List, batch_size=64, pad = None):
        """
        Embed large data batchwise as an iterator

        Same as the embed_batch function but as iterator, in case you want to process every batch embeddings before moving on to the next.
        (like writing to disk for very large data)
        :param sentences: List of sentences.
        :param batch_size: Size of the batches
        :param pad: (default: None) If pad is set all sentences will be padded (or cut repectively) to the desired length.
        :return: Yields the embeddings of a batch per iteration
        """
        ds = _temporal_dataset(sentences,)
        for b in tqdm(DataLoader(ds, batch_size=batch_size)):
            yield self.embed(b, pad=pad)



from .probability import Prob
from .lstm import LSTM, LSTMRD
from .label_layers import LabelEmbeddingScoring,AdaptiveCombination,LabelAttention, LabelSpecificSelfAttention
from .metric_layers import Bilinear, Metric
from .weighted_aggregation import AttentionWeightedAggregation
from .mogrifier import MogrifierLSTM, MogLSTM



from tqdm import tqdm
import torch
from ignite.metrics import Precision, Accuracy, Average
from ..metrics.multilabel import MultiLabelReport,AUC_ROC
from ..representation import is_transformer,get
from ..representation.labels import makemultilabels



class LanguageModelAbstract(torch.nn.Module):
    """
    Abstract class for Language Models. Defines fit, evaluate, predict and threshold methods for virtually any
    language model.
    Also provides a few default functions:
        _init_input_representations(): if self.representations exists, the default will load a embedding and corresponding tokenizer
        transform(): If self.tokenizer exists the default method wil use this to transform text into the models input

    """
    def __init__(self, loss=torch.nn.CrossEntropyLoss, optimizer=torch.optim.Adam, optimizer_params={"lr": 5e-5}, device="cpu", **kwargs):
        super(LanguageModelAbstract, self).__init__(**kwargs)

        self.device = device
        self.loss = loss
        self.optimizer = optimizer
        self.optimizer_params = optimizer_params
        self.PRECISION_DIGITS = 4

    def build(self):
        if isinstance(self.loss, type) and self.loss is not None:
            self.loss = self.loss().to(self.device)
        if isinstance(self.optimizer, type) and self.optimizer is not None:
            self.optimizer = self.optimizer(filter(lambda p: p.requires_grad, self.parameters()), **self.optimizer_params)
        self.to(self.device)

    def evaluate(self, data, batch_size=50):
        """
        Evaluation, return accuracy and loss
        """
        self.eval()  # set mode to evaluation to disable dropout
        correct = 0#p_1 = Accuracy()#Precision(is_multilabel=True,average=True)
        count = 0
        average_loss = Average()
        average_pp = Average()
        average_acc= Average()
        average_bpc= Average()
        entropy = torch.nn.CrossEntropyLoss(reduction="none").to(self.device)
        data_loader = torch.utils.data.DataLoader(data, batch_size=batch_size)

        with torch.no_grad():
            for i, b in enumerate(data_loader):
                count = i+1
                x = b[0].to(self.device)
                y = b[1].to(self.device)
                output = self(x)
                if hasattr(self, "regularize"):
                    l = self.loss(output, y) + self.regularize()
                else:
                    l = self.loss(output, y)

                average_loss.update(l.item())

                probabilities = torch.softmax(output, -1)
                bpc = (-probabilities * torch.log2(probabilities)).sum(-1)
                for i in bpc: average_bpc.update(i.item())
                # perplexity =torch.exp(entropy(output,y))#2**(- (torch.nn.functional.one_hot(y,probabilities.shape[-1]) * torch.log2(probabilities)).sum(-1))
                # for i in perplexity: average_pp.update(i.item())
                for i in (torch.max(output,-1)[1] == y).int(): average_acc.update(i)
        self.train()
        l = average_loss.compute()
        return {
            # "accuracy": accuracy.compute(),
            "valid_loss": round(l.item(), 2*self.PRECISION_DIGITS),
            "accuracy": round(average_acc.compute().item(),self.PRECISION_DIGITS),
            "perplexity": round(torch.exp(l).item(), 2*self.PRECISION_DIGITS),
            "BitsPer": round(average_bpc.compute().item(), 2*self.PRECISION_DIGITS)
        }

    def fit(self, train, valid = None, epochs=1, batch_size=16, valid_batch_size=50):
        self.train()
        validation=[]
        train_history = {"loss": []}
        for e in range(epochs):
            losses = {"loss": str(0.)}
            average = Average()
            train_loader = torch.utils.data.DataLoader(train, batch_size=batch_size, shuffle=True, pin_memory=True  )
            with tqdm(train_loader,
                      postfix=[losses], desc="Epoch %i/%i" %(e+1,epochs)) as pbar:
                for i, b in enumerate(train_loader):
                    self.optimizer.zero_grad()
                    x = b[0].to(self.device)
                    y = b[1].to(self.device)
                    output = self(x)
                    l = self.loss(output, y)
                    l.backward()
                    self.optimizer.step()
                    average.update(l.item())
                    pbar.postfix[0]["loss"] = round(average.compute().item(),2*self.PRECISION_DIGITS)
                    pbar.update()
                # torch.cuda.empty_cache()
                if valid is not None:
                    validation.append(self.evaluate(data=valid,
                                                    batch_size=valid_batch_size))
                    pbar.postfix[0].update(validation[-1])
                    pbar.update()
                # torch.cuda.empty_cache()
            train_history["loss"].append(average.compute().item())
        return{"train":train_history, "valid": validation }

    def generate(self, prompt="", steps=100, sample=True, beta=1):
        self.eval()
        with torch.no_grad():
            answer = self.encode(prompt)

            for _ in range(steps):
                cl = torch.tensor(answer[(-min(len(answer), self.max_len)):])[None,:]
                output = self(torch.LongTensor(cl).to(self.device))
                probabilities = torch.softmax(beta*output[0],-1).detach().cpu()
                if sample:
                    import numpy as np
                    next_token = np.random.choice(range(probabilities.shape[-1]),1, p=(probabilities/probabilities.sum()).numpy())[0]
                if not sample:
                    next_token=torch.argmax(probabilities).item()
                answer.append(next_token)
        self.train()
        return self.decode(answer)

    def transform(self, x):
        return torch.LongTensor([[self.tokenizer.token_to_id(t) for t in sequence] for sequence in x]).t()

    def num_params(self):
        total = sum(p.numel() for p in self.parameters())
        trainable = sum(p.numel() for p in self.parameters() if p.requires_grad)
        print("Parameters:\n"
              "Trainable:\t%i\n"
              "Fixed:\t%i\n"
              "-----------\n"
              "Total:\t%i" % (trainable, total-trainable,total))
    def representations(self, s):
        if not isinstance(s, list):
            s = [s]
        return self(self.transform(s).to(self.device), representations=True)



from .helpers import cooc_matrix
from .embeddings import get_nmf, get_node2vec, get_random_projection
from .graph_loaders import load_wordnet, load_wordnet_sample, load_NELL,load_elsevier, load_stw, load_nasa, \
    load_gesis, load_mesh
from .graph_operations import subgraphs, embed_align
from .graph_operations import plot_activation

register = {
    "wordnet": load_wordnet,
    "stw": load_stw,
    "nasa": load_nasa,
    "gesis": load_gesis,
    "mesh": load_mesh,
    "elsevier": load_elsevier

}

def get_graph(name: str):
    fct = register.get(name)
    if fct is None:
        raise FileNotFoundError
    else:
        return fct()

import networkx as nx
def get(name: [list, str]):
    if isinstance(name, str):
        name = [name]
    return nx.compose_all([get_graph(x) for x in name])





from setuptools import setup
import os
import subprocess


# Get Version from Pipeline, ignore leading 'v'
if os.environ.get('CI_COMMIT_TAG'):
    version = os.environ['CI_COMMIT_TAG'][1:]
elif os.environ.get('CI_JOB_ID'):
    version = os.environ['CI_JOB_ID']

# For local builds:
else:
    try:
        # Get latest git tag
        result = subprocess.run("git describe --tags", shell=True, stdout=subprocess.PIPE)
        version = result.stdout.decode('utf-8')[1:-1] + "-local"
    except:
        version = "local"


with open("mlmc/_version.py", "w") as f:
    f.write(f"__version__='{version}'")

setup(
    name='melmac',
    python_requires='>=3.6',
    version=version,
    packages=['mlmc',
              'mlmc.data',
              'mlmc.graph',
              'mlmc.layers',
              'mlmc.loss',
              'mlmc.metrics',
              'mlmc.models',
              'mlmc.representation'
              ],
    url='',
    license='',
    author='Janos Borst',
    author_email='borst@informatik.uni-leipzig.de',
    description='A package for neural multilabel and multiclass classification.',
    install_requires=['numpy',
                      'transformers',
                      'nltk',
                      'node2vec',
                      'scikit-learn',
                      'pytorch-ignite',
                      'tqdm',
                      'networkx',
                      'bs4',
                      'pytest',
                      'pytest-cov',
                      'rdflib',
                      'h5py'
                      ],
    include_package_data=True,
)



import numpy as np
from torch.nn.utils.rnn import pad_sequence
from torch import LongTensor

def charindex(query, maxlen, maxchar, alphabet):
    alphabet_map = dict(zip(alphabet, range(1, len(alphabet)+1)))
    result = np.full((len(query),maxlen,maxchar), 0., dtype="float32")
    for i, s in enumerate(query):
        indices=pad_sequence([LongTensor([alphabet_map.get(x, len(alphabet)+1)for x in word ]) for word in s.split()],batch_first=True).numpy()
        result[i, :min(indices.shape[-2],maxlen), :min(indices.shape[-1], maxchar)] = indices[:maxlen, :maxchar]
    return result
